import inspect
from collections.abc import Iterable, Sequence
from functools import cache
from itertools import product
from operator import itemgetter
from pathlib import Path
from typing import Any

import pytest
from logikal_browser import Browser
from logikal_browser.install import InstalledBrowser, installed_browsers
from logikal_browser.scenarios import Scenario, Settings

from pytest_logikal.utils import Fixture, Function

INSTALLED_BROWSERS_KEY = pytest.StashKey[dict[str, InstalledBrowser]]()


def pytest_sessionstart(session: pytest.Session) -> None:
    session.config.stash[INSTALLED_BROWSERS_KEY] = installed_browsers(
        install=not session.config.getoption('no_install'),
    )


def pytest_report_header(config: pytest.Config) -> str | list[str]:
    browsers = config.stash[INSTALLED_BROWSERS_KEY].values()
    if config.get_verbosity() > 0:
        return ['browsers:'] + [f'  {repr(browser.browser_version)}' for browser in browsers]
    return f'browsers: {', '.join(str(browser.browser_version) for browser in browsers)}'


def check_browser_fixture_request(request: Any) -> None:
    if (
        not hasattr(request, 'param')
        or 'installed_browser' not in request.param
        or 'settings' not in request.param
        or 'headless' not in request.param
    ):
        raise RuntimeError('You must specify a browser scenario for the browser fixture')


@pytest.fixture
def browser(tmp_path: Path, request: Any) -> Iterable[Browser]:  # noqa: D400, D402, D415
    """
    browser() -> logikal_browser.Browser

    Yield a scenario-specific :class:`~logikal_browser.Browser` sub-class instance.
    """
    check_browser_fixture_request(request)

    function = request.node.obj
    if not (source_file := inspect.getsourcefile(function)):  # pragma: no cover, defensive line
        raise RuntimeError(f'Source file cannot be found for "{function}"')
    source_path = Path(source_file)
    screenshot_path = source_path.parent / 'screenshots' / source_path.stem / function.__name__

    installed = request.param['installed_browser']
    with installed.browser_class(
        settings=request.param['settings'],
        version=installed.browser_version,
        headless=request.param['headless'],
        screenshot_path=screenshot_path,
        screenshot_tmp_path=tmp_path,
    ) as driver:
        yield driver


@cache
def get_installed_browsers() -> dict[str, InstalledBrowser]:
    return installed_browsers()


def set_browser(scenarios: Scenario | Sequence[Scenario], headless: bool = True) -> Fixture[Any]:
    """
    Apply the given scenarios to the :func:`~pytest_logikal.browser.browser` fixture.

    Args:
        scenarios: The scenarios to use.
        headless: Whether to use a headless browser instance.

    """
    def parametrized_test_function(function: Function) -> Any:
        argvalues = []
        installed = get_installed_browsers()
        for scenario in ([scenarios] if isinstance(scenarios, Scenario) else scenarios):
            browsers = scenario.browsers or installed.keys()
            settings = scenario.settings
            if isinstance(settings, Settings):
                settings = [settings]
            for browser_name, settings in product(browsers, settings):
                argvalues.append({
                    'name': f'{browser_name}_{settings.name}',
                    'installed_browser': installed[browser_name],
                    'settings': settings,
                    'headless': headless,
                })

        return pytest.mark.parametrize(
            argnames='browser', argvalues=argvalues, indirect=True, ids=itemgetter('name'),
        )(function)

    return parametrized_test_function



from contextlib import redirect_stdout
from io import StringIO

import pycodestyle
import pydocstyle
import pytest
from logikal_utils.project import tool_config

from pytest_logikal.file_checker import CachedFileCheckItem, CachedFileCheckPlugin
from pytest_logikal.plugin import ItemRunError


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('style')
    group.addoption('--style', action='store_true', default=False,
                    help='run pycodestyle and pydocstyle')


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line('markers', 'style: tests checking code and docstring style.')
    if config.option.style:
        config.pluginmanager.register(StylePlugin(config=config))


class StyleItem(CachedFileCheckItem):
    def run(self) -> None:
        messages = []

        # Run pycodestyle
        if not getattr(self.config.option, 'black', False):
            ignore = ['E133', 'W503']  # mutually exclusive checks with E123 and W504
            max_line_length = int(self.config.getini('max_line_length'))
            with redirect_stdout(StringIO()) as code_stdout:
                code_errors = pycodestyle.Checker(
                    filename=self.path,
                    max_line_length=max_line_length,
                    max_doc_length=max_line_length,
                    format='%(row)s:%(col)s: error: %(text)s (%(code)s)',
                    ignore=tool_config('pycodestyle').get('ignore', ignore),
                ).check_all()
            if code_errors:
                messages.append(code_stdout.getvalue().rstrip())

        # Run pydocstyle
        select = pydocstyle.violations.all_errors - {
            'D100', 'D101', 'D102', 'D103', 'D104', 'D105', 'D106', 'D107',
            'D200', 'D203', 'D204', 'D212', 'D406', 'D407', 'D408', 'D409',
        }
        select = tool_config('pydocstyle').get('select', select)
        if doc_errors := list(pydocstyle.check(filenames=[str(self.path)], select=select)):
            messages.extend(
                f'{error.line}: error: {error.short_desc} ({error.code})'
                if isinstance(error, pydocstyle.violations.Error) else f'error: {error}'
                for error in doc_errors
            )

        # Report errors
        if messages:
            raise ItemRunError('\n'.join(messages))


class StylePlugin(CachedFileCheckPlugin):
    name = 'style'
    item = StyleItem



import importlib
import string

import pytest
from license_expression import Licensing, Renderable, get_spdx_licensing
from logikal_utils.project import tool_config

from pytest_logikal.core import ReportInfoType
from pytest_logikal.plugin import Item, ItemRunError, Plugin

UNKNOWN_LICENSE = 'UNKNOWN'

# See https://spdx.org/licenses/
ALLOWED_LICENSES = [
    'Apache-2.0',
    'BSD-2-Clause',
    'BSD-3-Clause',
    'CC0-1.0',
    'ISC',
    'LGPL-2.0',
    'LGPL-2.0+',
    'LGPL-3.0',
    'LGPL-3.0+',
    'MIT',
    'MIT-0',
    'MIT-CMU',
    'MPL-2.0',
    'PSF-2.0',
    'Unlicense',
]
ALLOWED_PACKAGES = {
    'pylint': 'GPL-2.0-or-later',  # only used as a local tool
}

ALLOWED_LEGACY_LICENSES = [
    '3-Clause BSD License',
    'Apache 2.0',
    'Apache License 2.0',
    'Apache License Version 2.0',
    'Apache License, Version 2.0',
    'Apache Software License',
    'Apache-2.0',
    'BSD 3-Clause',
    'BSD License',
    'BSD',
    'BSD-2-Clause',
    'BSD-3-Clause',
    'CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
    'CMU License (MIT-CMU)',
    'GNU Lesser General Public License v2 (LGPLv2)',
    'GNU Lesser General Public License v2 or later (LGPLv2+)',
    'GNU Lesser General Public License v3 (LGPLv3)',
    'GNU Library or Lesser General Public License (LGPL)',
    'ISC License (ISCL)',
    'ISC',
    'MIT License',
    'MIT No Attribution License (MIT-0)',
    'MIT',
    'Mozilla Public License 2.0 (MPL 2.0)',
    'Public Domain',
    'Python Software Foundation License',
    'The MIT License (MIT)',
    'The Unlicense (Unlicense)',
    'Unlicense',
]

ALLOWED_LEGACY_PACKAGES = {
    'codespell': 'GPL-2.0-only',  # only used as a local tool
    'django-migration-linter': UNKNOWN_LICENSE,  # license is Apache License 2.0, see [1]
    'djlint': 'GNU General Public License v3 or later (GPLv3+)',  # only used as a local tool
    'facebook-business': 'LICENSE.txt',  # only used as a connector
    'facebook_business': 'LICENSE.txt',  # only used as a connector
    'html-tag-names': 'GNU General Public License v3 or later (GPLv3+)',  # local tool
    'html-void-elements': 'GNU General Public License v3 or later (GPLv3+)',  # local tool
    'pkg-resources': UNKNOWN_LICENSE,  # caused by an Ubuntu bug, see [2]
    'pkg_resources': UNKNOWN_LICENSE,  # caused by an Ubuntu bug, see [2]
    'pylint-django': 'GNU General Public License v2 or later (GPLv2+)',  # local plugin
    'pylint-plugin-utils': 'GNU General Public License v2 or later (GPLv2+)',  # local plugin
    'setuptools': UNKNOWN_LICENSE,  # see https://github.com/pypa/setuptools/issues/4956
}
# [1] https://github.com/3YOURMIND/django-migration-linter/issues/290
# [2] https://bugs.launchpad.net/ubuntu/+source/python-pip/+bug/1635463


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('licenses')
    group.addoption('--licenses', action='store_true', default=False, help='run license checks')


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line('markers', 'licenses: checks licenses.')
    if config.option.licenses:
        config.pluginmanager.register(LicensePlugin(config=config))


class LicenseItem(Item):
    @staticmethod
    def _normalize_label(label: str) -> str:
        # See https://packaging.python.org/en/latest/specifications/well-known-project-urls
        removal_map = str.maketrans('', '', string.punctuation + string.whitespace)
        return label.strip().translate(removal_map).lower()

    @staticmethod
    def _package_url(package: importlib.metadata.Distribution) -> str | None:
        # Parse project URLs
        urls: dict[str, str] = {}
        for project_url in package.metadata.get_all('project-url', []):
            label, url = project_url.split(',', 1)
            urls[LicenseItem._normalize_label(label)] = url.strip()

        # Check well-known labels
        # See https://packaging.python.org/en/latest/specifications/well-known-project-urls
        well_known_labels = [
            'homepage',
            'documentation', 'docs',
            'source', 'repository', 'sourcecode', 'github',
            'changelog', 'changes', 'whatsnew', 'history',
            'issues', 'bugs', 'issue', 'tracker', 'issuetracker', 'bugtracker',
        ]
        for label in well_known_labels:
            if url := urls.get(label):
                return url

        # Legacy home page URL, deprecated by PEP 753
        if url := package.metadata.get('home-page'):
            return url

        return None

    @staticmethod
    def _legacy_package_licenses(package: importlib.metadata.Distribution) -> list[str]:
        # Legacy licenses read from classifiers, deprecated by PEP 639
        if licenses := [
            classifier.split(' :: ')[-1]
            for classifier in package.metadata.get_all('classifier', [])
            if classifier.startswith('License ::') and classifier != 'License :: OSI Approved'
        ]:
            return licenses

        # Legacy license expression, deprecated by PEP 639
        if license_expression := package.metadata.get('license'):
            return [license_expression.replace('\n', ' ').replace('  ', ' ')]

        return [UNKNOWN_LICENSE]

    @staticmethod
    def _allowed_license(
        expression: str,
        allowed_expressions: list[Renderable],
        licensing: Licensing,
    ) -> bool:
        return any(
            licensing.contains(expression, allowed_expression)
            for allowed_expression in allowed_expressions
        )

    @staticmethod
    def _allowed_legacy_license(licenses: list[str], allowed_licenses: list[str]) -> bool:
        return any(
            allowed_license in (license, 'any')
            for license in licenses
            for allowed_license in allowed_licenses
        )

    def runtest(self) -> None:  # pylint: disable=too-many-locals
        config = tool_config('licenses')
        licensing = get_spdx_licensing()

        # SPDX licenses
        allowed_licenses = config.get('allowed_licenses', ALLOWED_LICENSES)
        allowed_packages = config.get('allowed_packages', ALLOWED_PACKAGES)
        allowed_licenses.extend(config.get('extend_allowed_licenses', []))
        allowed_packages.update(config.get('extend_allowed_packages', {}))

        allowed_license_expressions = [
            licensing.parse(allowed_license)
            for allowed_license in allowed_licenses
        ]
        allowed_package_license_expressions = {
            package: licensing.parse(allowed_license)
            for package, allowed_license in allowed_packages.items()
        }

        # Legacy licenses
        allowed_legacy_licenses = config.get('allowed_legacy_licenses', ALLOWED_LEGACY_LICENSES)
        allowed_legacy_packages = config.get('allowed_legacy_packages', ALLOWED_LEGACY_PACKAGES)
        allowed_legacy_licenses.extend(config.get('extend_allowed_legacy_licenses', []))
        allowed_legacy_packages.update(config.get('extend_allowed_legacy_packages', {}))

        warnings: list[str] = []
        for package in importlib.metadata.distributions():
            name = package.metadata['name']
            version = package.metadata['version']
            url = self._package_url(package) or f'https://pypi.org/project/{name}/'
            allowed_license = False
            allowed_package_license = False

            if license_expression := package.metadata.get('license-expression'):
                # SPDX license check
                parsed_license_expression = licensing.parse(license_expression)
                allowed_license = self._allowed_license(
                    expression=parsed_license_expression,
                    allowed_expressions=allowed_license_expressions,
                    licensing=licensing,
                )
                if package_license_expression := allowed_package_license_expressions.get(name):
                    allowed_package_license = self._allowed_license(
                        expression=parsed_license_expression,
                        allowed_expressions=[package_license_expression],
                        licensing=licensing,
                    )
            else:
                # Legacy license check
                licenses = self._legacy_package_licenses(package)
                license_expression = f'{'; '.join(licenses)} (legacy license definition)'
                allowed_license = self._allowed_legacy_license(
                    licenses=licenses,
                    allowed_licenses=allowed_legacy_licenses,
                )
                if package_license := allowed_legacy_packages.get(name):
                    allowed_package_license = self._allowed_legacy_license(
                        licenses=licenses,
                        allowed_licenses=[package_license],
                    )

            if (not allowed_license and not allowed_package_license):
                warnings.append(f'Warning: {license_expression}: {name}=={version} ({url})')

        if warnings:
            raise ItemRunError('\n'.join(warnings))

    def reportinfo(self) -> ReportInfoType:
        return (self.path, None, f'[{LicensePlugin.name}]')


class LicensePlugin(Plugin):
    name = 'licenses'
    item = LicenseItem



import zoneinfo
from collections.abc import Iterable, Sequence
from typing import Any, Protocol

import pytest
from django.conf import settings
from django.urls import reverse
from django.utils import timezone as django_timezone, translation
from factory import random as factory_random
from logikal_utils.random import DEFAULT_RANDOM_SEED
from mypy_django_plugin import config as mypy_django_plugin_config
from pytest_django.live_server_helper import LiveServer

from pytest_logikal.node_install import install_node_packages
from pytest_logikal.utils import Fixture, Function
from pytest_logikal.validator import Validator


def pytest_sessionstart(session: pytest.Session) -> None:
    if not session.config.getoption('no_install'):
        install_node_packages()

    css_or_svg = not session.config.getoption('no_css') or not session.config.getoption('no_svg')
    if not session.config.getoption('fast') and css_or_svg:
        # We start the validator service here to avoid pytest's output capturing
        Validator.service_url()


def pytest_configure(config: pytest.Config) -> None:
    # Patching django-stubs
    def parse_toml_file(self: Any, *_args: Any, **_kwargs: Any) -> None:
        self.django_settings_module = str(config.inicfg['DJANGO_SETTINGS_MODULE'])
        self.strict_settings = True

    mypy_django_plugin_config.DjangoPluginConfig.parse_toml_file = parse_toml_file  # type: ignore


@pytest.fixture(scope='session', autouse=True)
def faker_seed() -> int:
    """
    Set a default seed for Faker for deterministic testing. Automatically applied.
    """
    return DEFAULT_RANDOM_SEED


@pytest.fixture(scope='function', autouse=True)
def factory_seed() -> None:
    """
    Set a default seed for pytest-factoryboy for deterministic testing. Automatically applied.
    """
    factory_random.reseed_random(DEFAULT_RANDOM_SEED)


class LiveURL(Protocol):
    def __call__(
        self,
        name: str | None = None,
        args: Sequence[Any] | None = None,
        kwargs: dict[str, Any] | None = None,
    ) -> str:
        ...


@pytest.fixture
def live_url(live_server: LiveServer) -> LiveURL:  # noqa: D400, D402, D415, D417
    """
    live_url(name: str | None, args: Sequence[Any] | None, kwargs: dict[str, Any] | None) -> str

    Return the path to a URL.

    Args:
        name: The URL pattern name.
        args: The URL arguments to use.
        kwargs: The URL keyword arguments to use.

    """
    def live_url_path(
        name: str | None = None,
        args: Sequence[Any] | None = None,
        kwargs: dict[str, Any] | None = None,
    ) -> str:
        return live_server.url + (reverse(viewname=name, args=args, kwargs=kwargs) if name else '')
    return live_url_path


@pytest.fixture
def language(request: Any) -> Iterable[str]:
    """
    Return the currently activated language.
    """
    if language_code := getattr(request, 'param', None):
        with translation.override(language_code):
            yield language_code
    else:
        yield translation.get_language()


def set_language(*language_codes: str) -> Fixture[Any]:
    """
    Mark a test to run with each of the specified languages.

    Args:
        *language_codes: The language codes to use.

    .. note:: You must also use the :func:`language <pytest_logikal.django.language>` fixture in
        your test when applying this decorator.

    """
    def parametrized_test_function(function: Function) -> Any:
        return pytest.mark.parametrize(
            argnames='language', argvalues=language_codes, indirect=True,
            ids=lambda value: f'language={value}'
        )(function)
    return parametrized_test_function


def all_languages() -> Fixture[Any]:
    """
    Mark a test to run with every available language.

    .. note:: You must also use the :func:`language <pytest_logikal.django.language>` fixture in
        your test when applying this decorator.

    """
    return set_language(*(language_code for language_code, _ in settings.LANGUAGES))


@pytest.fixture
def timezone(request: Any) -> Iterable[str]:
    """
    Return the current time zone ID.
    """
    if zone_id := getattr(request, 'param', None):
        current_timezone = django_timezone.get_current_timezone()
        django_timezone.activate(zoneinfo.ZoneInfo(zone_id))
        yield zone_id
        django_timezone.activate(current_timezone)
    else:
        yield django_timezone.get_current_timezone_name()


def set_timezone(*zone_ids: str) -> Fixture[Any]:
    """
    Mark a test to run with each of the specified time zones.

    Args:
        *zone_ids: The time zone IDs to use.

    .. note:: You must also use the :func:`timezone <pytest_logikal.django.timezone>` fixture in
        your test when applying this decorator.

    """
    def parametrized_test_function(function: Function) -> Any:
        return pytest.mark.parametrize(
            argnames='timezone', argvalues=zone_ids, indirect=True,
            ids=lambda value: f'timezone={value}'
        )(function)
    return parametrized_test_function
